// bslmf_addconst.t.cpp                                               -*-C++-*-
#include <bslmf_addconst.h>

#include <bslmf_issame.h>

#include <bsls_bsltestutil.h>

#include <stdio.h>
#include <stdlib.h>

using namespace bsl;
using namespace BloombergLP;

//=============================================================================
//                                TEST PLAN
//-----------------------------------------------------------------------------
//                                Overview
//                                --------
// The component under test defines meta-functions 'bsl::add_const' and
// 'bsl::add_const_t', that add a top-level 'const'-qualifier to a template
// parameter type.  Thus, we need to ensure that the values returned by the
// meta-function are correct for each possible category of types.
//
// ----------------------------------------------------------------------------
// PUBLIC TYPES
// [ 1] bsl::add_const::type
// [ 1] bsl::add_const_t
//
// ----------------------------------------------------------------------------
// [ 2] USAGE EXAMPLE

// ============================================================================
//                     STANDARD BSL ASSERT TEST FUNCTION
// ----------------------------------------------------------------------------

namespace {

int testStatus = 0;

void aSsErT(bool condition, const char *message, int line)
{
    if (condition) {
        printf("Error " __FILE__ "(%d): %s    (failed)\n", line, message);

        if (0 <= testStatus && testStatus <= 100) {
            ++testStatus;
        }
    }
}

}  // close unnamed namespace

// ============================================================================
//               STANDARD BSL TEST DRIVER MACRO ABBREVIATIONS
// ----------------------------------------------------------------------------

#define ASSERT       BSLS_BSLTESTUTIL_ASSERT
#define ASSERTV      BSLS_BSLTESTUTIL_ASSERTV

#define LOOP_ASSERT  BSLS_BSLTESTUTIL_LOOP_ASSERT
#define LOOP0_ASSERT BSLS_BSLTESTUTIL_LOOP0_ASSERT
#define LOOP1_ASSERT BSLS_BSLTESTUTIL_LOOP1_ASSERT
#define LOOP2_ASSERT BSLS_BSLTESTUTIL_LOOP2_ASSERT
#define LOOP3_ASSERT BSLS_BSLTESTUTIL_LOOP3_ASSERT
#define LOOP4_ASSERT BSLS_BSLTESTUTIL_LOOP4_ASSERT
#define LOOP5_ASSERT BSLS_BSLTESTUTIL_LOOP5_ASSERT
#define LOOP6_ASSERT BSLS_BSLTESTUTIL_LOOP6_ASSERT

#define Q            BSLS_BSLTESTUTIL_Q   // Quote identifier literally.
#define P            BSLS_BSLTESTUTIL_P   // Print identifier and value.
#define P_           BSLS_BSLTESTUTIL_P_  // P(X) without '\n'.
#define T_           BSLS_BSLTESTUTIL_T_  // Print a tab (w/o newline).
#define L_           BSLS_BSLTESTUTIL_L_  // current Line number

//=============================================================================
//                  GLOBAL TYPEDEFS/CONSTANTS FOR TESTING
//-----------------------------------------------------------------------------

namespace {

struct TestType {
   // This user-defined type is intended to be used during testing as an
   // argument for the template parameter 'TYPE' of 'bsl::add_const'.
};

typedef void (TestType::*MethodPtrTestType) ();
    // This non-static function member type is intended to be used during
    // testing as an argument for the template parameter 'TYPE' of
    // 'bsl::add_const'.

typedef int TestType::* PMD;
    // This class public data member pointer type is intended to be used during
    // testing as an argument as an argument for the template parameter 'TYPE'
    // of 'bsl::add_const'.

}  // close unnamed namespace

//=============================================================================
//                              MAIN PROGRAM
//-----------------------------------------------------------------------------

int main(int argc, char *argv[])
{
    int                 test = argc > 1 ? atoi(argv[1]) : 0;
    bool             verbose = argc > 2;
    bool         veryVerbose = argc > 3;
    bool     veryVeryVerbose = argc > 4;
    bool veryVeryVeryVerbose = argc > 5;

    (void)veryVerbose;          // suppress warning
    (void)veryVeryVerbose;      // suppress warning
    (void)veryVeryVeryVerbose;  // suppress warning

    setbuf(stdout, NULL);       // Use unbuffered output

    printf("TEST " __FILE__ " CASE %d\n", test);

    switch (test) { case 0:
      case 2: {
        // --------------------------------------------------------------------
        // USAGE EXAMPLE
        //
        // Concerns:
        //: 1 The usage example provided in the component header file compiles,
        //:   links, and runs as shown.
        //
        // Plan:
        //: 1 Incorporate usage example from header into test driver, remove
        //:   leading comment characters, and replace 'assert' with 'ASSERT'.
        //:   (C-1)
        //
        // Testing:
        //   USAGE EXAMPLE
        // --------------------------------------------------------------------

        if (verbose) printf("\nUSAGE EXAMPLE\n"
                            "\n=============\n");

///Usage
///-----
// In this section we show intended use of this component.
//
///Example 1: Adding the 'const'-qualifier to A Type
///- - - - - - - - - - - - - - - - - - - - - - - - -
// Suppose that we want to add a 'const'-qualifier to a particular type.
//
// First, we create two 'typedef's -- a 'const'-qualified type ('MyConstType')
// and the same type without the 'const'-qualifier ('MyType'):
//..
        typedef int       MyType;
        typedef const int MyConstType;
//..
// Now, we add a 'const'-qualifier to 'MyType' using 'bsl::add_const' and
// verify that the resulting type is the same as 'MyConstType':
//..
        ASSERT(true == (bsl::is_same<bsl::add_const<MyType>::type,
                                                         MyConstType>::value));
//..
// Finally, if the current compiler supports alias templates C++11 feature, we
// add a 'const'-qualifier to 'MyType' using 'bsl::add_const_t' and verify that
// the resulting type is the same as 'MyConstType':
//..
#ifdef BSLS_COMPILERFEATURES_SUPPORT_ALIAS_TEMPLATES
        ASSERT(true ==
                 (bsl::is_same<bsl::add_const_t<MyType>, MyConstType>::value));
#endif  // BSLS_COMPILERFEATURES_SUPPORT_ALIAS_TEMPLATES
//..

      } break;
      case 1: {
        // --------------------------------------------------------------------
        // 'bsl::add_const::type'
        //   Ensure that the 'typedef' 'type' of 'bsl::add_const' has the
        //   correct type for a variety of template parameter types.
        //
        //   Ensure that the 'bsl::add_const_t' represents the return type of
        //   'bsl::add_const' meta-function.
        //
        // Concerns:
        //: 1 'bsl::add_const' adds a top-level 'const'-qualifier only to
        //:   primitive types, pointer types, and user-defined types.
        //:
        //: 2 'bsl::add_const' does not add a 'const'-qualifier to reference
        //:   types, function types, or types that are already
        //:   'const'-qualified.
        //:
        //: 3 'bsl::add_const_t' represents the return type of 'bsl::add_const'
        //:   meta-function for a variety of template parameter types.
        //
        // Plan:
        //  1 Verify that 'bsl::add_const::type' has the correct type for each
        //    concern. (C1-2)
        //
        //  2 Verify that 'bsl::add_cons_t' has the same type as the return
        //    type of 'bsl::add_const' for a variety of template parameter
        //    types. (C-3)
        //
        // Testing:
        //   bsl::add_const::type
        //   bsl::add_const_t
        // --------------------------------------------------------------------

        if (verbose) printf("\n'bsl::add_const::type'\n"
                            "\n======================\n");

        // C-1
        ASSERT((is_same<add_const<int>::type, int const>::value));
        ASSERT((is_same<add_const<int *>::type, int * const>::value));
        ASSERT((is_same<add_const<int const *>::type,
                                                   int const * const>::value));
        ASSERT((is_same<add_const<int[]>::type, const int[]>::value));
        ASSERT((is_same<add_const<int[2]>::type, const int[2]>::value));
        ASSERT((is_same<add_const<int[][2]>::type, const int[][2]>::value));
        ASSERT((is_same<add_const<TestType>::type, TestType const>::value));

        ASSERT((is_same<add_const<void>::type, const void>::value));
        ASSERT((is_same<add_const<volatile void>::type,
                                                 const volatile void>::value));

        // C-2
        ASSERT((is_same<add_const<int &>::type, int &>::value));
        ASSERT((is_same<add_const<TestType &>::type, TestType &>::value));
        ASSERT((is_same<add_const<int (int)>::type, int (int)>::value));
        ASSERT((is_same<add_const<int const>::type, int const>::value));

#ifdef BSLS_COMPILERFEATURES_SUPPORT_ALIAS_TEMPLATES

        if (verbose) printf("\n'bsl::add_const_t'\n"
                            "\n==================\n");

        // C-3
        ASSERT((is_same<add_const  <int        >::type,
                        add_const_t<int        >>::value));
        ASSERT((is_same<add_const  <int *      >::type,
                        add_const_t<int *      >>::value));
        ASSERT((is_same<add_const  <int const *>::type,
                        add_const_t<int const *>>::value));
        ASSERT((is_same<add_const  <TestType   >::type,
                        add_const_t<TestType   >>::value));

        ASSERT((is_same<add_const  <int &     >::type,
                        add_const_t<int &     >>::value));
        ASSERT((is_same<add_const  <TestType &>::type,
                        add_const_t<TestType &>>::value));
        ASSERT((is_same<add_const  <int (int) >::type,
                        add_const_t<int (int) >>::value));
        ASSERT((is_same<add_const  <int const >::type,
                        add_const_t<int const >>::value));

        ASSERT((is_same<add_const  <MethodPtrTestType   >::type,
                        add_const_t<MethodPtrTestType   >>::value));
        ASSERT((is_same<add_const  <MethodPtrTestType & >::type,
                        add_const_t<MethodPtrTestType & >>::value));
        ASSERT((is_same<add_const  <MethodPtrTestType * >::type,
                        add_const_t<MethodPtrTestType * >>::value));
        ASSERT((is_same<add_const  <MethodPtrTestType *&>::type,
                        add_const_t<MethodPtrTestType *&>>::value));

        ASSERT((is_same<add_const<PMD   >::type, add_const_t<PMD   >>::value));
        ASSERT((is_same<add_const<PMD & >::type, add_const_t<PMD & >>::value));
        ASSERT((is_same<add_const<PMD * >::type, add_const_t<PMD * >>::value));
        ASSERT((is_same<add_const<PMD *&>::type, add_const_t<PMD *&>>::value));

#endif  // BSLS_COMPILERFEATURES_SUPPORT_ALIAS_TEMPLATES
      } break;
      default: {
        fprintf(stderr, "WARNING: CASE `%d' NOT FOUND.\n", test);
        testStatus = -1;
      }
    }

    if (testStatus > 0) {
        fprintf(stderr, "Error, non-zero test status = %d.\n", testStatus);
    }

    return testStatus;
}

// ----------------------------------------------------------------------------
// Copyright 2013 Bloomberg Finance L.P.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
// ----------------------------- END-OF-FILE ----------------------------------
